#ifndef __CLF_GEOMETRY__
#define __CLF_GEOMETRY__

#include <vector>
#include <list>

float	const	pi	= 3.1415926535897932384626433832795f;
float	const	tol = 0.0001f;		// float type tolerance 

inline	float	DegreesToRadians(float deg) {
	return deg * pi / 180.0f;
}

inline	float	RadiansToDegrees(float rad) {	
	return rad * 180.0f / pi;
}

inline int dti(double d) {   // approximate float to integer
	return (int)floor(d+0.5);
}

// angle in DEG !!!
float myCos(float a);
float mySin(float a);
float myTan(float a);
void init_math_lib();

//////////////////////////////////////////////////////////////////////////////////
/////////////// 2D section

class clf_point2D{
public:
    float x;
    float y;
    clf_point2D():x(0.f), y(0.f){
    }
    clf_point2D(float _x, float _y):x(_x), y(_y){
    }
    clf_point2D(int _x, int _y):x((float)_x), y((float)_y){
    }
    ~clf_point2D(){
    }
};

class clf_rect2D{
public:
    float left;
    float top;
    float right;
    float bottom;
    clf_rect2D() : left(0.f),top(0.f),right(0.f),bottom(0.f) {
    }
    clf_rect2D(float l, float t, float r, float b) : left(l),top(t),right(r),bottom(b) {
    }
    clf_rect2D(clf_rect2D& p) : left(p.left),top(p.top),right(p.right),bottom(p.bottom) {
    }
    ~clf_rect2D(){
    }
};

class clf_polygon2D : public std::vector< clf_point2D>{
public:
	clf_rect2D bounds;
	clf_polygon2D(){
	}
    ~clf_polygon2D(){
	}

	void add_point(const clf_point2D& pt){
		if (size() == 0)
		{
			bounds = clf_rect2D(pt.x, pt.y, pt.x, pt.y);
		}
		else
		{
			if (bounds.left > pt.x) bounds.left = pt.x;
			if (bounds.right < pt.x) bounds.right = pt.x;
			if (bounds.top < pt.y) bounds.top = pt.y;
			if (bounds.bottom > pt.y) bounds.bottom = pt.y;
		}
		push_back(clf_point2D(pt.x, pt.y));
	}
};

class clf_triangle2D
{
public:
	clf_triangle2D(){}
	clf_triangle2D(const clf_triangle2D& t){
		verts[0] = t.verts[0];
		verts[1] = t.verts[1];
		verts[2] = t.verts[2];
	}
	~clf_triangle2D(){}

	clf_point2D verts[3];
};

bool intersect(const clf_point2D& p1, const clf_point2D& p2, const clf_point2D& p3, const clf_point2D& p4, clf_point2D* p0);
bool point_in_polygon(clf_polygon2D& poly, clf_point2D pt);
bool polygons_collide(clf_polygon2D& poly1, clf_polygon2D& poly2);
void rotate2D(float cx, float cy, float *x, float *y, float angle);
// check for intersection, but exclude the vertices
bool intersect_no_vertex(const clf_point2D& p1, const clf_point2D& p2, const clf_point2D& p3, const clf_point2D& p4, clf_point2D* p0);
// check if the line segment (p1,p2) is inside a polygon
bool line_in_polygon(clf_polygon2D& poly, clf_point2D& p1, clf_point2D& p2);
bool triangulate(clf_polygon2D& poly, std::vector<clf_triangle2D>& result);

//////////////////////////////////////////////////////////////////////////////////
/////////////// 3D section

class clf_matrix4x4;

class clf_vector3D{
public:
	float x,y,z;
	clf_vector3D(){
		x = y = z = 0;
	}
	clf_vector3D(float _x, float _y, float _z):x(_x), y(_y), z(_z){
	}
	clf_vector3D(clf_vector3D& v):x(v.x), y(v.y), z(v.z){
	}
	clf_vector3D& operator=(clf_vector3D& v){
		x = v.x; y = v.y; z = v.z;
		return *this;
	}
	clf_vector3D& operator*=(float s){	// scalar multiply
		x *= s; y *= s; z *= s;
		return *this;
	}
	clf_vector3D& operator/=(float s){	// scalar divide
		x /= s; y /= s; z /= s;
		return *this;
	}
	clf_vector3D operator-(void) {
		return clf_vector3D(-x, -y, -z);
	}
	operator float* () {
		return &x;
	}
	clf_vector3D& scale(float s){
		x *= s; y *= s; z *= s;
		return *this;
	}
	float length(){
		return sqrt(x*x + y*y + z*z);
	}
	void normalize(){ 
		float fLength = 1.0f / length();
		scale(fLength);
	}
	clf_vector3D inverse(){
		clf_vector3D r(-x, -y, -z);
		return r;
	}
	clf_vector3D& rotate(clf_matrix4x4& mat);

};

inline clf_vector3D operator + (clf_vector3D& m1, clf_vector3D& m2){
	clf_vector3D r(m1);
	r.x += m2.x; r.y += m2.y; r.z += m2.z;
	return r;
}
inline clf_vector3D operator - (clf_vector3D& m1, clf_vector3D& m2){
	clf_vector3D r(m1);
	r.x -= m2.x; r.y -= m2.y; r.z -= m2.z;
	return r;
}
inline float dotProduct(const clf_vector3D& v1, const clf_vector3D& v2){
	return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z; 
}
inline clf_vector3D crossProduct(const clf_vector3D& v1, const clf_vector3D& v2){
	clf_vector3D vResult;
	vResult.x = v1.y*v2.z - v2.y*v1.z;
	vResult.y = -v1.x*v2.z + v2.x*v1.z;
	vResult.z = v1.x*v2.y - v2.x*v1.y;
	return vResult;
}
inline clf_vector3D operator*(clf_vector3D& v1, float s)
{	// scalar multiply
	clf_vector3D r(v1);
	r.x *= s; r.y *= s; r.z *= s;
	return r;
}													 
inline clf_vector3D operator*(float s, clf_vector3D& v1)
{	// scalar multiply
	clf_vector3D r(v1);
	r.x *= s; r.y *= s; r.z *= s;
	return r;
}													 

class clf_vector4 {
	float data[4];
public:
	clf_vector4(){}
	clf_vector4(const clf_vector4& m1){
		memcpy(data, m1.data, 4*sizeof(float));
	}
	clf_vector4(float x, float y, float z,float a){
		data[0] = x;
		data[1] = y;
		data[2] = z;
		data[3] = a;
	}
	virtual ~clf_vector4(){}
	virtual float& operator[](int index){
		return data[index];
	}
	virtual float operator[](int index) const{
		return data[index];
	}
	clf_vector4& operator = (clf_vector4& m1){
		memcpy(data, m1.data, 4*sizeof(float));
		return *this;
	}
	operator float* () {
		return data;
	}
	clf_vector4& scale(float sc){
		for (int i=0; i<4; ++i)
			data[i] *= sc;
		return *this;
	}
	clf_vector4 operator + (clf_vector4& m1){
		clf_vector4 r;
		for (int i=0; i<4; ++i)
			r[i] = data[i] + m1.data[i];
		return r;
	}
	clf_vector4 operator - (clf_vector4& m1){
		clf_vector4 r;
		for (int i=0; i<4; ++i)
			r[i] = data[i] - m1.data[i];
		return r;
	}
};

class clf_matrix4x4 {
	float data[16];
	float DetIJ(int i, int j);
public:
	clf_matrix4x4(){
		memset(data,0,sizeof(float)*16);
	}
	clf_matrix4x4(const clf_matrix4x4& m1){
		memcpy(data, m1.data, 16*sizeof(float));
	}
	virtual ~clf_matrix4x4(){}
	virtual float& operator[](int index){
		return data[index];
	}
	virtual float operator[](int index) const{
		return data[index];
	}
	clf_matrix4x4& operator = (clf_matrix4x4& m1){
		memcpy(data, m1.data, 16*sizeof(float));
		return *this;
	}
	operator float* () {
		return data;
	}
	clf_matrix4x4& scale(float sc){
		for (int i=0; i<16; ++i)
			data[i] *= sc;
		return *this;
	}
	clf_matrix4x4 operator + (clf_matrix4x4& m1){
		clf_matrix4x4 r;
		for (int i=0; i<16; ++i)
			r[i] = data[i] + m1.data[i];
		return r;
	}
	clf_matrix4x4 operator - (clf_matrix4x4& m1){
		clf_matrix4x4 r;
		for (int i=0; i<16; ++i)
			r[i] = data[i] - m1.data[i];
		return r;
	}
	clf_matrix4x4 operator *(float sc){
		clf_matrix4x4 r;
		for (int i=0; i<16; ++i)
			r[i] = data[i] * sc;
		return r;
	}
	clf_vector4 operator *(clf_vector4& v);

	void loadIdentity();
	clf_matrix4x4 operator *(const clf_matrix4x4& m2);
	void Transpose();
	clf_matrix4x4& Invert();
	void TranslationMatrix(float x, float y, float z);
	void ScalingMatrix(float x, float y, float z);
	void RotationMatrix(float angle, float x, float y, float z);
	void RotationMatrix(float angle, clf_vector3D& vAxis);
	void MakeShadowMatrix(clf_vector3D vPoints[3], clf_vector4& vLightPos);
};

void clf_getNormalVector(clf_vector3D& vP1, clf_vector3D& vP2, clf_vector3D& vP3, clf_vector3D& vNormal);
void clf_rotateVector(const clf_vector3D& vSrcVector, const clf_matrix4x4& mMatrix, clf_vector3D& vOut);
void clf_unrotateVector(const clf_vector3D& vSrcVector, const clf_matrix4x4& mMatrix, clf_vector3D& vOut);
/*
dVector dMatrix::UnrotateVector (const dVector &v) const
{
	return dVector (v % m_front, v % m_up, v % m_right);
}
*/

void clf_transformPoint(const clf_vector3D& vSrcVector, const clf_matrix4x4& mMatrix, clf_vector3D& vOut);
void clf_getPlaneEquation(clf_vector3D& vPoint1, clf_vector3D& vPoint2, clf_vector3D& vPoint3, clf_vector4& vPlane);
float clf_distanceToPlane(clf_vector3D& vPoint, clf_vector4& vPlane);

void rotate_point(float x, float y, float z, clf_vector3D& point);
void rotate_point_rad(float x, float y, float z, clf_vector3D& point);
void rotate_f(float x, float y, float z, clf_vector3D& point);


class clf_quaternion {
	float r;
	float i,j,k;
public:
	clf_quaternion() : r(1), i(0), j(0), k(0){
	}
	clf_quaternion(float _r, float _i, float _j, float _k) : r(_r), i(_i), j(_j), k(_k){
	}
	clf_quaternion(clf_quaternion& q) : r(q.r), i(q.i), j(q.j), k(q.k){
	}

	float magnitude(){
		return (float) sqrt(r*r + i*i + j*j + k*k);
	}
};



////////////////////////////////////////////////////////////////////

#endif // __CLF_GEOMETRY__
